; FREE32.ASM 10/9/2011
;  Based on FREETEST.ASM by Eric Auer
;
; compile:		nasm -o free32.com free32.asm
; (nasm is a free open source Assembler: http://nasm.sf.net/)
;
; syntax:		FREE32 [d:] [?]
; d: is the [optional] drive selection (default is current drive)
; ? displays a help message
; 2/22/2011 - added code to calculate and display values after
;  the decimal point for Kbytes and MBytes
; 2/26/2011 - modified code to use EAX:EDX instead of AX:DX
;  to allow full bytes printouts up to 2 TB.
; 3/08/2011 - corrected issue with floppy disks
; 3/13/2011 - slightly revised format of results
; 8/23/2011 - converted to all 32-bit
; 9/27/2011 - added MByte output
; 9/28/2011-10/2/2011 - minor output format improvements
; 10/7-9/2011 - added calculations of %free and %used; format changes
; 2012-11-09 - jmh: change names to KiBytes & MiBytes; added /Q; return %free;
;  read the country info for the thousands & decimal separators

%define FAT32 1 	; (define after adding the int 21.7303 call!)

	org 100h	; a COM program

start:	mov ax,2524h	; install critical error handler, no need to save
	mov dx,i24handler	; previous one: int 22-24 get restored
	int 21h 		; at program exit.

	mov ax,3800h	; get current country info
	mov dx,country
	int 21h
	jc .nosep
	mov al,[country+7]
	mov [tsep],al
	mov al,[country+9]
	mov [dsep],al
.nosep:
	mov ah,19h	; get current drive
	int 21h
	add al,'A'	; this call returns 0 for A:, 1 for B: ...
	mov [drvnam],al ; set drive letter

%if FAT32
	mov ax,7300h	; get a FAT32 property
	mov dl,0	; current drive
	mov cl,1	; "dirty-buffers flag"
	int 21h
	cmp ax,7300h	; AX still unchanged?
	jz oldkernel
	; ignore the actual results of int 21.7300 (AL, AH)
	mov byte [kern32],1	; kernel has FAT32 support

oldkernel:
%endif

	xor bp,bp	; drive, 1 for A:, 2 for B: ... or 0 for "current"

	mov si,81h	; command line arguments
cline:	mov ax,[si]	; 2 chars at a time
	inc si
	cmp al,13	; end?
	jz main
	cmp al,'?'
	jz help
	cmp al,' '	; ignore space and control characters
	jbe cline
	cmp al,'/'	; test slashes...
	jz opt
	cmp al,'-'	; ...and dashes
	jz opt
	cmp ah,':'	; is the FOLLOWING char a colon?
	jnz help	; else there is a syntax error
	cmp al,'a'	; is it lower case?
	jb nlower
	sub al,'a'-'A'	; then make it upper case!
nlower: cmp al,'A'
	jb help
	cmp al,'Z'
	ja help
	inc si		; skip over the already parsed colon!
	mov [drvnam],al ; update drive letter
	sub al,'A'-1	; turn A: into 1, B: into 2, etc.
	mov bp,ax	; store drive number
	and bp,31	; zap extra bits
	jmp short cline

opt:	cmp ah,'q'
	je .q
	cmp ah,'Q'
	jne help
.q:	mov byte [quiet],1
	inc si
	jmp short cline

help:	mov al,[tsep]
	mov [hsep1],al
	mov [hsep2],al
	mov dx,helpmsg
	mov ah,9	; show message
	int 21h
	mov ax,4c00h	; return errorlevel 0
	int 21h

i24handler:
	mov al,3	; auto-answer "abort/retry/fail" question
	iret		; with "fail". No prompt will be shown.

main:	mov al,[drvnam] ; fetch selected drive letter
	mov [drvchr],al ; put it into our message

	mov al,[dsep]
	mov [dot_total],al
	mov [dot_mtotal],al
	mov [dot_diff],al
	mov [dot_mdiff],al
	mov [alloc_pc+4],al
	mov [dot_free],al
	mov [dot_mfree],al
	mov [free_pc+4],al

%if FAT32
	cmp byte [kern32],0
	jz usefat16
	mov al,[drvnam] ; drive letter
	mov [what32],al ; plug letter into "x:\",0 string
	mov dx,what32	; in DS: string
	mov di,size32	; in ES: size info buffer
	mov cx,30h	; buffer size, at least 2ch
	mov ax,7303h	; get extended free space
	int 21h
	jc bug2
	cmp al, 0
	je bug2 	; error?
			; Else: dd [size32+8]	= bytes per sector
			;	dd [size32+14h] = free sectors
	mov ax,[di+10]
	or ax,ax
	jnz bug2	; sector size > 64k not supported
	mov ax,[di+8]	; sector size
; ---	cmp ax,512
; ---	jb bug2 	; sector size < 512 not supported

	mov [sector],ax ; save sector size for translation
	mov bx,[di+4]	; sectors per cluster
	mov [sectors],bx ; save sectors/cluster
	mul bx		; calc cluster size
	mov [cluster],ax ; save clust_bytes for translation

	mov eax,[di+20] ; free sector count, low and high
	xor edx,edx

	push di 	; save DI

; save free sectors for difference calculation
savfree:
	mov    [dfree],eax	; save free sectors

; calculate free bytes

	call mul512e

	jc bug2

	call store_free

	pop di		; recall di value

	mov eax,[di+24] ; total sector count, low and high
	xor edx,edx

; save total sectors for difference calculation
savtot:

	mov    [dtotal],eax

	call mul512e

	jc bug2

	call store_total

; determine allocated bytes by difference

;calcdiff:
	clc
	mov    eax,[dtotal]	; calculate difference between
	sub    eax,[dfree]	; total sectors and sectors allocated
	xor edx,edx
	mov [ddiff],eax 	; store allocated value

	call mul512e

	jc bug2

	call store_diff

	call calc_pc

	jmp done

bug2:	jmp bug

usefat16:
%endif
	mov ah,36h	; get free disk space (old MS DOS 2+ style)
	mov dx,bp	; DL is the drive number (0 default, 1 A:, 2 B: ...)
	int 21h
	cmp ax,-1	; error?
	jz bug
	mov [BX_val],bx ; save free clusters
	mov [DX_val],dx ; save total clusters
	push dx 	; save total clusters
	mov [sectors],ax ; save sectors/cluster
	mov [sector],cx  ; save sector size
	xor dx,dx	; prepare for 32bit mul
	mul cx		; multiply AX (sectors per cluster) by CX (bytes per sector)
	mov [cluster],ax ; save bytes/cluster
	push ax 	; save cluster bytes
	mul bx		; multiply result by BX (number of free clusters)

; save free bytes for difference calculation
	std
	mov    di,dfree+2
	xchg ax,dx
	stosw
	xchg ax,dx
	stosw

	mov eax,[dfree]
	xor edx,edx

	call store_free

	pop ax		; retrieve cluster bytes
	pop bx		; retrieve total clusters (in BX)
	xor dx,dx	; prepare for 32bit mul
	mul bx		; multiply by number of total clusters

; save total bytes for difference calculation
	std
	mov    di,dtotal+2
	xchg ax,dx
	stosw
	xchg ax,dx
	stosw

	mov eax,[dtotal]
	xor edx,edx

	call store_total

;calcdiff:
	clc
	mov    eax,[dtotal]   ;calculate difference between
;	 mov	dx,[dtotal+2] ;  total bytes and free bytes
	sub    eax,[dfree]
;	 sbb	dx,[dfree+2]
	xor edx,edx

	mov [ddiff],eax       ; store difference value

	call store_diff

	call calc_pc

done:	cmp byte [quiet],0
	jne exit

	mov ax,[sector]   ; retrieve sector size
	xor dx,dx

	call conv_32

	mov di,sectora+7 ; storage for printout
	call   TRANSLATE

	mov ax,[sectors]   ; retrieve sectors/cluster
	xor dx,dx

	call conv_32

	mov di,sectorsa+3 ; storage for printout
	call   TRANSLATE

	mov ax,[cluster]   ; retrieve bytes/cluster
	xor dx,dx

	call conv_32

	mov di,clustera+9 ; storage for printout
	call   TRANSLATE

	mov dx,crlf	; announce size info
	mov ah,9	; show message
	int 21h

exit:	mov al,[pcfree] ; return %free
	mov ah,4ch
	int 21h

bug:	cmp byte [quiet],0
	jne .exit
	mov dx,nomsg	; problems encountered (e.g. no such FAT drive)
	mov ah,9
	int 21h
.exit:	mov ax,4c00h	; return 0 (error)
	int 21h

store_free:
	mov    di,drvmsg2+18	 ;storage for ascii printout
	call   TRANSLATE

	push eax	; save bytes
	push edx

	call div1024e	 ; convert to Kbytes

	mov    di,free_k+16    ;storage for ascii printout
	call   TRANSLATE

	mov di,dot_free+3
	call get_dots

	pop edx 	; recall bytes
	pop eax

	call div1024sq		; calculate free MBytes

	mov    di,free_m+9	; storage for ascii printout
	call   TRANSLATE

	mov di,dot_mfree+3
	call get_dots1

	ret

store_total:
	mov    di,total+18     ;storage for ascii printout
	call   TRANSLATE

	push eax	; save bytes
	push edx

	call div1024e	 ; convert to Kbytes

	mov    di,total_k+16	;storage for ascii printout
	call   TRANSLATE

	mov di,dot_total+3
	call get_dots

	pop edx 	; recall bytes
	pop eax

	call div1024sq		 ; calculate total MBytes

	mov    di,total_m+9	 ; storage for ascii printout
	call   TRANSLATE

	mov di,dot_mtotal+3
	call get_dots1
	ret

store_diff:
	mov    di,diff+18     ;storage for ascii printout
	call   TRANSLATE

	push eax	; save bytes
	push edx

	call div1024e	 ; convert to Kbytes

	mov    di,diff_k+16    ;storage for ascii printout
	call   TRANSLATE

	mov di,dot_diff+3
	call get_dots

	pop edx 	; recall bytes
	pop eax

	call	div1024sq      ; calculate occupied MBytes

	mov    di,diff_m+9     ;storage for ascii printout
	call   TRANSLATE

	mov di,dot_mdiff+3
	call get_dots1

	ret

calc_pc:
	mov eax,[dfree] 	; recall free value
	mov ebx,100		; prepare for multiply by 100
	xor edx,edx
	mul ebx 		; multiply by 100
	mov ebx,[dtotal]	; prepare to divide by dtotal
	div ebx 		; divide by dtotal
	mov [pcfree],al 	; preserve value for ERRORLEVEL
	push edx		; save remainder
	xor edx,edx		; clear EDX

	mov di,free_pc+3	; store ascii printout of quotient
	call TRANSLATE

	pop eax 		; recall remainder (in EAX)
	mov ebx,100
	xor edx,edx
	mul ebx 		; multiply remainder by 100
	mov ebx,[dtotal]
	div ebx 		; divide 100*remain by dtotal
	shr ebx,1		; divide dtotal by 2
	cmp edx,ebx		; is remainder > dtotal/2?
	jb no_r1		; if not, skip rounding
	inc eax 		; if so, round up
no_r1:	xor edx,edx		; clear EDX

	mov di,free_pc+6	; store ascii printout
	call TRANSLATE

	mov eax,[ddiff] 	; recall used value
	mov ebx,100		; prepare for multiply by 100
	xor edx,edx
	mul ebx 		; multiply by 100
	mov ebx,[dtotal]	; prepare to divide by dtotal
	div ebx 		; divide by dtotal
	push edx		; save remainder
	xor edx,edx		; clear remainder

	mov di,alloc_pc+3	 ; store ascii printout of quotient
	call TRANSLATE

	pop eax 		; recall remainder (in EAX)
	mov ebx,100
	xor edx,edx
	mul ebx 		; multiply remainder by 100
	mov ebx,[dtotal]
	div ebx 		; divide 100*remain by dtotal
	shr ebx,1		; divide dtotal by 2
	cmp edx,ebx		; is remainder > dtotal/2?
	jb no_r2		; if not, skip rounding
	inc eax 		; if so, round up
no_r2:	xor edx,edx		; clear EDX

	mov di,alloc_pc+6	 ; store ascii printout
	call TRANSLATE

	ret

conv_32:
	std
	mov    di,ax_dx+2
	xchg ax,dx
	stosw
	xchg ax,dx
	stosw

	mov eax,[ax_dx]
	xor edx,edx
	cld

	ret

div1024e:
	mov ebx,1024	; divide by 1,024
	div ebx
	mov [remain32],edx
	xor edx,edx
	ret

div1024sq:
	mov ebx,1048576 ; divide bytes by 1,048,576
	div ebx
	mov [remain32],edx ; save remainder
	xor edx,edx	; clear remainder from EDX
	ret

get_dots:	; convert remainder to ASCII
	std
	mov eax,[remain32]
	cmp eax,0	; is it zero?
	jnz chk512	; if not, goto chk512
	cmp word [sectors],1	; is it 1 sector/cluster?
	jz no_dot	; if eax is zero, but 1 sector/cluster, goto no_dot
	sub di,2	; else if zero, but not 1 sector/cluster
	mov al," "	;  then blank ".0"
	stosb
	stosb
	jmp short no_dot ; and skip all the rest

chk512:
	cmp eax,512	 ; special case
	jnz nextdot

	push di
	sub di,2
	mov al,"5"	; replace "0" with "5"
	stosb
	pop di
	jmp short no_dot

nextdot:
	xor edx,edx
	mov ebx,1000	 ; multiply remainder by 1,000
	mul ebx

; then divide by 1,024
	mov ebx,1024	 ; divide by 1,024
	div ebx
	cmp edx,512	 ; remainder < 512?
	jb no_round
	inc eax 	 ; round up if edx = 512 or more
no_round:
	xor edx,edx

	call TRANSLATE	; convert to decimal ASCII

no_dot:
	ret

get_dots1:	; convert remainder to ASCII
	std
	mov eax,[remain32]
	cmp eax,0	 ; is it zero?
	jz no_dot1	 ; if so, skip all the rest

	xor edx,edx
	mov ebx,1000	 ; multiply remainder by 1,000
	mul ebx

; then divide by 1,048,576 (1,024^2)
	mov ebx,1048576   ; divide by 1,048,576
	div ebx
	cmp edx,524288	  ; remainder < 524,288? (1/2 of divisor)
	jb no_round1
	inc eax 	 ; round up if edx = 524,288 or more
no_round1:
	xor edx,edx

	call TRANSLATE	; convert to decimal ASCII
	cmp eax,9	; is it more than 9?
	ja no_dot1	; if so, skip
	std
	mov al,'0'
	stosb
	cld

no_dot1:
	ret

mul512e:
; assume 512 sector size
	mov cx,9	; multiply by 512
againe: shl eax,1	; by shifting bits
	rcl edx,1
	loop againe
	ret

TRANSLATE:
	std
	push eax	 ; save EAX and EDX for kbyte conversion
	push edx
	XCHG	EAX,EDX
	MOV	EBX,10		 ;Convert to decimal
	PUSH	EBP		 ;Use BP for temporary storage
	XOR	CX,CX		;Zero CL counter

NEXT_COUNT:
	CMP	CL,3		; is it 3?
	JNZ	NO_COMMA	; if not, go to NO_COMMA
	push ax
	MOV	AL,[tsep]	; put comma in AL
	STOSB			; store it
	pop ax
	XOR	CL,CL		; and clear counter

NO_COMMA:
	MOV	EBP,EDX 	  ; save current DX value in BP
	XOR	EDX,EDX 	  ; clear DX
	DIV	EBX		 ; divide by 10
	XCHG	EAX,EBP 	  ; recall current DX value from BP
				; and save current AX value in BP
	DIV	EBX		 ; divide by 10
	XCHG	EAX,EDX
	ADD	AL,'0'		;Convert to ASCII
	STOSB			; store result
	INC	CL		; increment counter
	MOV	EAX,EBP 	  ; recall current AX value from BP
	OR	EBP,EDX 	  ; are we done?
	JNZ	NEXT_COUNT	; If not, continue

	POP	EBP
	pop edx 	 ; recall EAX and EDX
	pop eax
	cld
	RET

nomsg	db "Cannot determine free space.",13,10,"$"

helpmsg db 13,10,"Free disk space checker by L R Holliday 10/9/2011 & Jason Hood 2012-11-09",13,10
	db 13,10
	db "Usage:",13,10
	db "  FREE32 [X:] [/Q] [?]",13,10
	db "Drive spec is optional, default is the current drive.",13,10
	db "/Q prevents output, to just return the free percentage.",13,10
	db 13,10
	db "Example: FREE32 C:",13,10,13,10
	db "Note: KiBytes = Bytes/1"
hsep1	db ",024; MiBytes = KiBytes/1"
hsep2	db ",024",13,10,13,10
%if FAT32
	db "FAT32 drives are fully supported by FREE32.",13,10
%else
	db "This FREE32 ignores free space above 2 GiBytes.",13,10
%endif
	db "$"

drvnam	db "?:\",0	; fails if drive letter not given, so use int 21.19

crlf	db 13,10
drvmsg	db "Drive "
drvchr	db "?:"
	db "                 Bytes"
	db "          KiBytes"
	db "           MiBytes",13,10
	db "              "
	db "  "
	db "  "
	db 13,10,"Total Space: "
total	times 19 db " "  ; 19 spaces for ASCII results
total_k times 17 db " "  ; 17 spaces for ASCII results
dot_total db ".0  "
total_m times 10 db " "  ; 10 spaces for ASCII results
dot_mtotal db ".0  "

	db 13,10," Used Space: "
diff	times 19 db " "
diff_k	times 17 db " "
dot_diff db ".0  "
diff_m	times 10 db " "
dot_mdiff db ".0   "
alloc_pc  db "    .0  %"

	db 13,10," Free Space: "
drvmsg2 times 19 db " "
free_k	times 17 db " "
dot_free db ".0  "
free_m	times 10 db " "
dot_mfree db ".0   "
free_pc   db "    .0  %"
	db 13,10,13,10

sectorsa db "     Sectors/Cluster"
clustera db "           Bytes/Cluster",13,10,36
sectora  db "         Bytes/Sector",13,10,36

sectors dw 0 ; sectors/cluster
sector	dw 0 ; sector size
cluster dw 0 ; bytes/cluster

BX_val	dw 0 ; free cluster value
DX_val	dw 0 ; total cluster value
ax_dx	dd 0 ; temp storage
remain	dw 0
remain32  dd 0

dfree	dd 0  ; space to store hex free space results
dtotal	dd 0  ; also total space
ddiff	dd 0  ; also allocated space

quiet	db 0  ; /Q
pcfree	db 0  ; exit value
tsep	db ',' ; thousands separator
dsep	db '.' ; decimal separator

%if FAT32

kern32	db 0		; flag: set to 1 if kernel supports FAT32
what32	db "Q:\",0	; the drive which we want to have checked

	; filled by int 21.7303.dsdx=&drvnam.esdi=&sizeinf.cx=2ch
	; which returns ax=7300 for non-FAT32 DOS versions or carry set
	; if FAT32 is supported but an error (AX) occurred during the call.
size32	dw 30h	; returns size
	dw 0	; returns version
%if 0	; rest of the data can be in uninitialized (BSS) space in RAM...
	dd 0,0,0,0	; sec/clust1, by/clust, free clusters1, total clusters
	dd 0,0,0,0	; free sect2, total sect2, free clust2, total clust2
	dd 0,0		; reserved
%endif
%endif
country: ;resb 34
